local FRACUNIT = FRACUNIT //Fast constant access

// To prevent duplicate freeslots
local function CheckSlot(item) // this function deliberately errors when a freeslot does not exist
	if _G[item] == nil // this will error by itself for states and objects
		error() // this forces an error for sprites, which do actually return nil for some reason
	end
end

local function SafeFreeslot(...)
	local item = ...
	if item == nil
		return
	end

	if pcall(CheckSlot, item)
		print("\131NOTICE:\128 " .. item .. " was not allocated, as it already exists.")
	else
		freeslot(item)
	end

	SafeFreeslot(select(2, ...))
end

local function valid(mo)
	return mo and mo.valid
end

-- a local version of string.split for us to use
local function castlesplit(str, sep)
    local fields = {}
	if not str then return fields end

    local sep = sep or " "
    local pattern = string.format("([^%s]+)", sep)
    string.gsub(str, pattern, function(c) fields[#fields + 1] = c end)

    return fields
end

//Axises for the Maids and Crystals
local CastleAxises = {}
local CastleCrystals = {}
local CastleFloatingCrystals = {}
addHook("NetVars", function(network)
	CastleAxises = network($)
	CastleCrystals = network($)
	CastleFloatingCrystals = network($)
end)

/*
States + Thinker for the crystal objects.
Use with the Toad stuff below.
*/
SafeFreeslot("SPR_ELV0", "S_CASTLE_WALLCRYSTAL", "S_CASTLE_FLOATCRYSTAL")
states[S_CASTLE_WALLCRYSTAL] = {SPR_ELV0, R_Char2Frame("h")|FF_FULLBRIGHT, -1}
states[S_CASTLE_FLOATCRYSTAL] = {SPR_ELV0, R_Char2Frame("m")|FF_FULLBRIGHT, -1}

local function PrepareCrystal(crystal)
	if crystal.state == S_CASTLE_FLOATCRYSTAL
		table.insert(CastleFloatingCrystals, crystal)
	else
		table.insert(CastleCrystals, crystal)
	end

	//Spawn the glow
	local glow = P_SpawnMobjFromMobj(crystal, 0,0,0, MT_OVERLAY)
	glow.target = crystal
	glow.state = S_OILLAMPFLARE
	glow.frame = $ | FF_ADD | FF_FULLBRIGHT
	glow.color = crystal.color
	glow.colorized = true
	if crystal.state == S_CASTLE_FLOATCRYSTAL
		glow.spriteyoffset = 56*FRACUNIT
	else
		glow.spriteyoffset = 32*FRACUNIT
	end
	crystal.hnext = glow

	crystal.crystalz = crystal.z
	crystal.crystalframe = crystal.frame
end

//Crystals are processed in a ThinkFrame rather than a MobjThinker to improve performance
addHook("ThinkFrame", function()
	if not mapheaderinfo[gamemap].aetherscastle then return end

    //Set the frame and fullbright based on that frame
	//It's the same for all the crystals, so we calculate
	//it only once per tic to get a performance boost
    local frame = abs(sin(leveltime * (ANG10 >> 2))) / (FRACUNIT/6)
    frame = min(max($, 0), 4)

	//Same deal here, this value is identical for all crystals
	//so no need to recalculate it for every iteration
	local crystalz = (cos(leveltime * (ANG10 >> 1)) + FRACUNIT)*8

	local CastleCrystals = CastleCrystals
    for i = 1, #CastleCrystals do
        local crystal = CastleCrystals[i]
        if crystal and crystal.valid
            crystal.frame = crystal.crystalframe + frame
        end
    end

	local CastleFloatingCrystals = CastleFloatingCrystals
	for i = 1, #CastleFloatingCrystals do
        local crystal = CastleFloatingCrystals[i]
        if crystal and crystal.valid
            crystal.frame = crystal.crystalframe + frame
            crystal.z = crystal.crystalz + crystalz * (crystal.scale / FRACUNIT)
        end
    end
end)

/*
Castle Interaction Object, starts a dialog when you press jump when touching it.
Don't be afraid to make the radius way bigger than the actual object to simulate
a large radius of effect!

Dialog IDs are determined by UDMF Arguments 1-7, and will progress through them
each time you interact with a specific object.
*/

SafeFreeslot("MT_CASTLE_INTERACTABLE",
"S_CASTLE_INTERACT1", "S_CASTLE_INTERACT2", "S_VIO_DANCE")
mobjinfo[MT_CASTLE_INTERACTABLE] = {
	spawnhealth = 1000,
	spawnstate = S_INVISIBLE,
	radius = 96*FRACUNIT,
	height = 32*FRACUNIT,
	flags = MF_SPECIAL|MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_SCENERY,
}

A_CastleInteract = function(mo)
	mo.angle = leveltime*(ANGLE_11hh*2/3)
end
states[S_CASTLE_INTERACT1] = {SPR_NULL, 0, 1, nil, 0, 0, S_CASTLE_INTERACT2}
states[S_CASTLE_INTERACT2] = {SPR_ELV0, F|FF_PAPERSPRITE, 1, A_CastleInteract, 0, 0, nil}
states[S_VIO_DANCE] = {SPR_ELV0, T|FF_ANIMATE, -1, nil, (Y - T), 3, S_VIO_DANCE}

//Touching an interactable object, let's try and interact with it
addHook("TouchSpecial", function(mobj, mo)
	local player = mo.player
	local localplayer = (player == consoleplayer or player == secondarydisplayplayer)
	local args = mobj.castleargs or (mobj.tracer and mobj.tracer.castleargs)
	if player.bot then return true end //No bots!
	if not P_IsObjectOnGround(mo) then return true end //Need to be grounded
	if not (player.panim >= PA_IDLE and player.panim <= PA_WALK)
	and mo.state != S_PLAY_SKID then return true end //Need to be in a walking state
	if mo.castleinteract < 0 then return true end //In the middle of conversation

	//Pressing jump starts a dialog
	if mo.castleinteract == 1
		//Get the dialog progress for this specific player
		local var = "castledialog"
		if not netgame
			var = $ + "0"
		else
			var = $ + #player
		end
		mobj[var] = ($ and $ < 7 and $ + 1) or 1

		//Get the next dialog
		local dialog = args[mobj[var] - 1]
		if not dialog
			mobj[var] = 1
			dialog = args[0]
			if not dialog then return true end
		end

		//Play the dialog out for the player
		mo.castleinteract = -1
		mo.castleinteractor = mobj
		mobj.target = mo
		A_ForceStop(mo)
		if localplayer
			startmindtext(dialog,
			(mobj.tracer and mobj.tracer.valid and mobj.tracer.color) or mobj.color)
		end

	//Spawn an indicator for the local player
	else
		if localplayer
			P_SpawnLockOn(player, mobj, S_CASTLE_INTERACT1)
		end
		player.pflags = $ | PF_JUMPDOWN
	end

	return true
end, MT_CASTLE_INTERACTABLE)

//Jump detector for above
//If it's negative, that means we're in a conversation and must move again
//before we're allowed to start another.
addHook("PlayerThink", function(player)
	if not mapheaderinfo[gamemap].aetherscastle then return end
	local mo = player.mo
	if not valid(mo) then return end

	if mo.castleinteract == nil
		mo.castleinteract = 0
	elseif mo.castleinteract < 0
		if mo.castleinteract != -1
		and (player.cmd.buttons & (BT_JUMP|BT_SPIN)
		or player.cmd.forwardmove or player.cmd.sidemove)
			mo.castleinteract = BT_JUMP
		elseif valid(mo.castleinteractor)
			player.drawangle = R_PointToAngle2(mo.x, mo.y, mo.castleinteractor.x, mo.castleinteractor.y)
		end
		mo.castleinteract = $ - 1
	elseif player.cmd.buttons & BT_JUMP
		mo.castleinteract = $ + 1
	elseif mo.castleinteract
		mo.castleinteract = 0
	end
end)

/*
Maids in Aether's Castle
These gals walk within any NiGHTS Axis points they're placed in, and can walk
between ones that are touching as well.
Make a string of Axis points to create a pathing structure!
*/

//Get the nearest axis
local function NearestCastleAxis(mo)
	Silverhorn.perfStats:start("Maid axis searcher")

	local closestaxis, closestdist = nil, INT32_MIN
	local minx, maxx = INT32_MIN, INT32_MAX
	local mox, moy = mo.x, mo.y

	for i = 1, #CastleAxises do
		local axis = CastleAxises[i]
		local x, y = axis.x, axis.y

		//Preliminary test to avoid useless function calls
		if x < minx or x > maxx then continue end

		//Accurate test when actually needed
		local dist = R_PointToDist2(x, y, mox, moy)
		if dist >= closestdist then continue end

		closestaxis, closestdist = axis, dist
		minx, maxx = x - dist, x + dist
	end

	Silverhorn.perfStats:stop("Maid axis searcher")
	return closestaxis
end

//Maid defintions
SafeFreeslot("MT_CASTLE_MAID", "S_CASTLE_MAID_STND", "S_CASTLE_MAID_WALK")
mobjinfo[MT_CASTLE_MAID] = {
	spawnhealth = 1000,
	spawnstate = S_CASTLE_MAID_STND,
	seestate = S_CASTLE_MAID_WALK,
	radius = mobjinfo[MT_PLAYER].radius,
	height = mobjinfo[MT_PLAYER].height*11/8,
	speed = FRACUNIT*4,
	reactiontime = TICRATE, //Min Random Wait Time
	mass = TICRATE*4, //Max Random Wait Time
	damage = 5, //How many steps can we ignore boundaries
	painchance = MT_BOXSPARKLE, //Sparkle effect object
	flags = MF_SOLID|MF_PUSHABLE|MF_RUNSPAWNFUNC|MF_SCENERY,
}
states[S_CASTLE_MAID_STND] = {SPR_ELV0, R_Char2Frame("r")|FF_ANIMATE, -1, nil, 3, 4, S_CASTLE_MAID_STND}
states[S_CASTLE_MAID_WALK] = {SPR_ELV0, R_Char2Frame("v")|FF_ANIMATE, -1, nil, 3, 2, S_CASTLE_MAID_WALK}

addHook("MobjThinker", function(mo)
	//Silverhorn.perfStats:start("Maid thinker")

	//Spawn threshold
	if not mo.threshold
		//Get a Castle Interactable if this maid can be talked to
		if mo.castleargs and mo.castleargs[0]
			mo.tracer = P_SpawnMobjFromMobj(mo, 0,0,0, MT_CASTLE_INTERACTABLE)
			mo.tracer.height = mo.height
			mo.tracer.tracer = mo
		end
		mo.flags2 = $ | MF2_SLIDEPUSH
	else
		//Set a random accessible color if we don't have one
		if not mo.color
			local colors = {}
			for i = 1, #skincolors - 1
				if skincolors[i].accessible then table.insert(colors, i) end
			end
			mo.color = colors[P_RandomKey(#colors + 1)] or SKINCOLOR_GREEN
		end

		//Set default shadow
		if not mo.shadowscale then mo.shadowscale = FRACUNIT end
	end

	//Increase Timer
	if mo.state == S_CASTLE_MAID_WALK
		mo.threshold = $ + 2
	else
		mo.threshold = $ + 1
	end

	//Float effect
	mo.spriteyoffset = (cos(mo.threshold * (ANG10 >> 1)) + FRACUNIT)*4
	//P_SpawnMobjFromMobj(mo, (P_RandomFixed() - INT16_MAX)*32,
	//(P_RandomFixed() - INT16_MAX)*32,
	//mo.spriteyoffset, mo.info.painchance).momz = -mo.scale*P_MobjFlip(mo)

	//Castle Interactable stuff
	local tracer = mo.tracer
	if tracer and tracer.valid
		//Stop and face whoever we're talking to
		local target = tracer.target
		if target and target.valid and target.castleinteractor == tracer
		and target.castleinteract and target.castleinteract < 0
			if mo.state != S_CASTLE_MAID_STND
				mo.state = S_CASTLE_MAID_STND
			end
			mo.reactiontime = mo.info.reactiontime //Max Random Wait Time
			mo.angle = R_PointToAngle2(mo.x, mo.y, target.x, target.y)
		end

		P_MoveOrigin(tracer, mo.x, mo.y, mo.z)
		tracer.z = $ + FixedMul(mo.scale, mo.spriteyoffset)
	end

	//Movement logic
	mo.reactiontime = $ - 1
	if mo.reactiontime <= 0
		//Randomize wait time
		mo.reactiontime = P_RandomRange(mo.info.reactiontime, mo.info.mass)
		mo.movefactor = mo.info.damage

		//Start stand
		if mo.state == S_CASTLE_MAID_WALK
			mo.state = S_CASTLE_MAID_STND

		//Start walk
		else
			mo.maidaxis = NearestCastleAxis(mo)

			//Walk towards the nearest axis
			if mo.cusval
				local axis = mo.maidaxis
				if axis and axis.valid
					mo.angle = R_PointToAngle2(mo.x, mo.y, axis.x, axis.y)
				end
				mo.angle = $/ANGLE_90*ANGLE_90
			else
				//Randomize our walking angle
				if P_RandomChance(FRACUNIT/2)
					mo.angle = $/ANGLE_90*ANGLE_90
				else
					mo.angle = P_RandomRange(1, 4)*ANGLE_90
				end

				//Check our movement to see if we'd run into something or outside
				//the nearest axis, turn if yes
				local turn = P_RandomChance(FRACUNIT/2) and 1 or -1
				for i = 1, 4
					local pos = {x = mo.x, y = mo.y, z = mo.z}
					local speed = FixedMul(mo.info.speed, mo.scale)*mo.reactiontime
					local movement = P_TryMove(mo,
					mo.x + P_ReturnThrustX(mo, mo.angle, speed),
					mo.y + P_ReturnThrustY(mo, mo.angle, speed),
					false)
					//local axis = NearestCastleAxis(mo)

					if not movement //or (axis and axis.valid
					//and R_PointToDist2(mo.x, mo.y, axis.x, axis.y) > axis.radius)
						P_SetOrigin(mo, pos.x, pos.y, pos.z)
						mo.angle = $ + ANGLE_90*turn
					else
						P_SetOrigin(mo, pos.x, pos.y, pos.z)
						break
					end
				end
			end
			mo.state = S_CASTLE_MAID_WALK
		end

	//During walk, go forward
	elseif mo.state == S_CASTLE_MAID_WALK
		local speed = FixedMul(mo.info.speed, mo.scale)
		local movement = P_TryMove(mo,
		mo.x + P_ReturnThrustX(mo, mo.angle, speed),
		mo.y + P_ReturnThrustY(mo, mo.angle, speed),
		false)

		//Reset movement timer if we run into a wall or outside an axis
		if mo.movefactor
			mo.movefactor = $ - 1
		elseif not movement
			mo.reactiontime = 0
		elseif not mo.cusval
			local axis = mo.maidaxis
			if axis and axis.valid
			and R_PointToDist2(mo.x, mo.y, axis.x, axis.y) > axis.radius
				mo.reactiontime = 0
				mo.cusval = 1 //Walk towards the nearest axis
			end
		elseif mo.cusval and mo.reactiontime == 1
			mo.cusval = 0 //Reset axis walk
		end
	end
	//Silverhorn.perfStats:stop("Maid thinker")
end, MT_CASTLE_MAID)

/*
Use Toad in Aether's Castle and UDMF arguments to spawn any object we want.

--------------------------------------------------------------------------------

String Argument 1 determines what object type Toad will transform into.
It can also determine the object flags to use instead, separated by |

Examples:
MT_BLUECRAWLA
MF_SOLID|MF_PUSHABLE|MF_NOGRAVITY

--------------------------------------------------------------------------------

String Argument 2 determines what visuals or state Toad will use, separated by |

Example:
S_MARIOBUSH1|SPR_KOOP|B|FF_FULLBRIGHT|FF_ADD|SKINCOLOR_RED

--------------------------------------------------------------------------------

Argument 8 - 10 determine the radius, height, and shadow size respectively.
Shadow size is NOT the same as shadowscale. It means directly how big the shadow
is in FRACUNITs, and is not influenced by the radius at all.

Examples:
Argument 8 = 16
Argument 9 = 48
Argument 10 = 16

--------------------------------------------------------------------------------

For whatever reason, sometimes Arguments 6-10 just won't save properly in
Aether's Castle. If this happens, just make a fresh new Toad.
*/

local ToadTransform = function(mo, mt)
	if not mapheaderinfo[gamemap].aetherscastle or not mt then return end
	local objecttype = mt.stringargs[0]
	local stateinfo = mt.stringargs[1]
	local thinker = false
	local changedtype = false
	local oldheight = mo.height

	local args = {}
	for i = 0, 9
		args[i] = mt.args[i]
	end

	//Transform Toad into a different object
	if objecttype and #objecttype > 3
	and string.sub(objecttype, 1, 3) == "MT_"
		if not pcall(function() objecttype = _G[$] end)
			mo.state = S_INVISIBLE
			return
		end
		mo.type = objecttype

		//Delete all Maids, reduces lag
		/*
		if mo.type == MT_CASTLE_MAID
			mo.fuse = 1 //Delete the OG
			return
		end
		*/

		local info = mobjinfo[objecttype]
		mo.radius = FixedMul(info.radius, mo.scale)
		mo.height = FixedMul(info.height, mo.scale)
		A_SetObjectFlags(mo, info.flags)
		mo.health = (info.spawnhealth and info.spawnhealth or 1)
		mo.reactiontime = info.reactiontime
		mo.dispoffset = info.dispoffset

		if mo.flags & MF_RUNSPAWNFUNC
			mo.state = info.spawnstate
		else
			P_SetMobjStateNF(mo, info.spawnstate)
		end
		thinker = true
		changedtype = true

	//Give Toad certain flags
	elseif objecttype and #objecttype > 3
	and string.sub(objecttype, 1, 3) == "MF_"
		local flags = 0
		local split = castlesplit(objecttype, "|")
		for i, flag in ipairs(split)
			flags = $ | (_G[flag] or 0)
		end
		A_SetObjectFlags(mo, flags)
		thinker = true
	end

	//Change Toad's state, sprite, and frame
	local split = castlesplit(stateinfo, "|")
	for i, state in ipairs(split)
		//Change Toad's state
		if state and string.sub(state, 1, 2) == "S_"
			mo.state = _G[state]
			if mo.tics != -1 or states[mo.state].action != nil
				thinker = true
			end

		//Change Toad's sprite
		elseif state and string.sub(state, 1, 4) == "SPR_"
			mo.sprite = _G[state]

		//Change Toad's frame flags
		elseif state and string.sub(state, 1, 3) == "FF_"
			local flag = _G[state]
			mo.frame = $ | flag

			if flag == FF_FLOORSPRITE
				mo.renderflags = $ | RF_NOSPLATBILLBOARD
			end

		//Change Toad's color
		elseif state and string.sub(state, 1, 10) == "SKINCOLOR_"
			mo.color = _G[state]

		//Change Toad's frame number
		elseif state and #state == 1 and state != " " and R_Char2Frame(state)
			mo.frame = ($ &~ FF_FRAMEMASK) | R_Char2Frame(state)

		elseif tonumber(state)
			mo.frame = ($ &~ FF_FRAMEMASK) | tonumber(state)
		end
	end

	//Change Toad's radius, height, and shadowscale
	if args[7] then mo.radius = args[7]*mo.scale end
	if args[8] then mo.height = args[8]*mo.scale end
	mo.shadowscale = FixedDiv(args[9]*FRACUNIT, FixedDiv(mo.radius, mo.scale))

	//Reverse gravity height fix
	if mo.eflags & MFE_VERTICALFLIP and mo.height != oldheight
		mo.z = $ - (mo.height - oldheight)
	end

	//Add a dialog object if we have dialog args and didn't change the type
	if not changedtype and args[0]
		local interact = P_SpawnMobjFromMobj(mo, 0,0,0, MT_CASTLE_INTERACTABLE)
		interact.height = mo.height
		interact.tracer = mo
		A_CapeChase(interact, 1)
	end

	//If possible, replace Toad with an object without a thinker
	mo.castleargs = args
	if not thinker
		local mobj = P_SpawnMobjFromMobj(mo, 0,0,0, MT_BLUEBERRYBUSH)
		mobj.radius = mo.radius
		mobj.height = mo.height
		mobj.z = mo.z
		mobj.shadowscale = mo.shadowscale
		mobj.state = mo.state
		mobj.sprite = mo.sprite
		mobj.frame = mo.frame
		mobj.castleargs = args

		mo.fuse = 1 //Delete the OG

	elseif changedtype
		return
	end

	//Get rid of the object type just in case a mod has custom behavior
	mo.type = MT_RAY
end
addHook("MapThingSpawn", ToadTransform, MT_TOAD)

//Spawn in one random object at a Fang Waypoint
addHook("MapLoad", function(map)
	if not mapheaderinfo[map].aetherscastle then return end

	local waypoints = {}
	CastleAxises = {}
	CastleCrystals = {}
	for mo in mobjs.iterate()
		if mo.type == MT_FANGWAYPOINT then table.insert(waypoints, mo) end
		if mo.type == MT_AXIS then table.insert(CastleAxises, mo) end
        if mo.state == S_CASTLE_WALLCRYSTAL or mo.state == S_CASTLE_FLOATCRYSTAL then
			PrepareCrystal(mo)
        end
	end

	/*
	for i, mo in ipairs(waypoints)
		if valid(mo) and valid(mo.spawnpoint)
			ToadTransform(mo, mo.spawnpoint)
		end
	end
	*/

	///*
	local mo = waypoints[P_RandomKey(#waypoints) + 1]
	if valid(mo) and valid(mo.spawnpoint)
		ToadTransform(mo, mo.spawnpoint)
	end
	//*/
end)