--[[
* customsave_io.lua
* (sprkizard)
* (June 11, 2020 16:08)
* Desc: Library for Saving/Loading data into a custom file in SRB2 I/O
		https://wiki.srb2.org/wiki/User:Rapidgame7/iodocs

* Usage: Refer to - https://github.com/sprkizard/srb2-lua-lib/wiki/I-O-Easy-Custom-Save-Files
]]

-- Converts a Lua table to a text representation.
local function Serialize(input)
	if type(input) == "string" then
		return "[[" .. input:gsub("](%--)]", "]-%1]") .. "]]"
	elseif type(input) == "number" then
		return "[" .. input .. "]"
	elseif type(input) == "boolean" then
		if input then
			return "T"
		else
			return "F"
		end
	elseif type(input) == "nil" then
		return "N"
	elseif type(input) == "table" then

		local str = "{"
		
		local delimit = false
		for k,v in pairs(input) do
			if delimit then
				str = $ .. ";"
			end
			
			str = $ .. Serialize(k) .. ":" .. Serialize(v)
			
			delimit = true
		end
		
		str = $ .. "}"
		return str
		
	else
		error("table contains a non-serializable element!")
	end
end

-- Converts the text representation of a Lua table back to the original table.
local function Deserialize(input)
	local output

	if input:sub(1, 2) == "[[" then -- string
		local endpos = input:find("]]")
		output = input:sub(3, endpos-1):gsub("]%-(%--)]", "]%1]")
		input = $:sub(endpos+2)
	elseif input:sub(1, 1) == "[" then -- number
		local endpos = input:find("]")
		output = tonumber(input:sub(2, endpos-1))
		input = $:sub(endpos+1)
	elseif input:sub(1, 1) == "T" then -- true
		output = true
		input = $:sub(2)
	elseif input:sub(1, 1) == "F" then -- false
		output = false
		input = $:sub(2)
	elseif input:sub(1, 1) == "N" then -- nil
		output = nil
		input = $:sub(2)
	elseif input:sub(1, 2) == "{}" then -- empty table
		output = {}
		input = $:sub(3)
	elseif input:sub(1, 1) == "{" then -- table
		output = {}
		
		while input:sub(1, 1) ~= "}" do
			local key, value
			
			key, input = Deserialize(input:sub(2))
			if input:sub(1, 1) ~= ":" then
				error("string is badly formatted! "..input)
			end
			
			value, input = Deserialize(input:sub(2))
			
			output[key] = value
		end
		
		input = $:sub(2)
	else
		error("string is badly formatted! "..input)
	end
	
	if input:len() then
		return output, input
	else
		return output
	end
end




-- Simple encrtyps/decrypts a text string or file
local function convert(chars,dist,inv)
	local charInt = string.byte(chars);
	for i=1,dist do
		if(inv)then charInt = charInt - 1; else charInt = charInt + 1; end
		if(charInt<32)then
			if(inv)then charInt = 126; else charInt = 126; end
		elseif(charInt>126)then
			if(inv)then charInt = 32; else charInt = 32; end
		end
	end
	return string.char(charInt);
end

local function crypt(str,k,inv)
	local enc= "";
	for i=1,#str do
		if(#str-k[5] >= i or not inv)then
			for inc=0,3 do
				if(i%4 == inc)then
					enc = enc .. convert(string.sub(str,i,i),k[inc+1],inv);
					break;
				end
			end
		end
	end
	if(not inv)then
		for i=1,k[5] do
			enc = enc .. string.char(P_RandomRange(32,126));
		end
	end
	return enc;
end

local enc1 = {1,2,3,4,0};
local enc2 = {124,533,663,123,27};






-- Reading data

-- Checks if a local file exists
function Silverhorn.I_FileExists()

	local f = io.openlocal(Silverhorn.LocalDataPath, "r")
	
	if f ~= nil then
		f:close()
		return true
	else
		return false
	end
end

-- TODO: deleteme
-- Reads a saved table into a variable from a file path
function Silverhorn.LoadSave2(fn)

	local encrypt = true
	io.open(Silverhorn.LocalDataPath, "r", function(file, filename)
	
		if (file == nil) then
			return
		else
			local loaded_file = file:read("*a")
			local ret = ""

			if encrypt then
				ret = Deserialize(crypt(loaded_file, enc2, true))
			else
				ret = Deserialize(loaded_file)
			end

			fn(ret)
		end
	
	end)

end

function Silverhorn.LoadSave()
	
	local file = io.openlocal(Silverhorn.LocalDataPath) -- TODO: try regular open()
	
	if (file == nil) then
		-- file:close()
	else
		local loaded_file = file:read("*a")

		file:close()

		return Deserialize(crypt(loaded_file, enc2, true))
	end
	
	return false
end



-- Writing data

-- Writes a save table to the specified file path
function Silverhorn.WriteSave()
	
	local file = io.openlocal(Silverhorn.LocalDataPath, "w+") -- TODO: try regular open()
	file:write(crypt(Serialize(Silverhorn.LocalData), enc2))

	file:flush()
	file:close()
end

-- Saves a value to a specific field of the save data
function Silverhorn.SaveValue(section, field, value)
	if Silverhorn.LocalData[section] and Silverhorn.LocalData[section][field] != nil
		Silverhorn.LocalData[section][field] = value
		//Silverhorn.WriteSave() // Lach: instead of writing every time we update a value, let's just save when the game closes
	else
		print("\x83 WARNING:\x80 Failed to save Silverhorn Save Data.")
		print("Either section " + section + " or field " + field + " does not exist.")
	end
end

-- Silverhorn Gamedata
-- Probably better off keeping this here instead of CONFIG for ease of use

Silverhorn.LocalData = {}
Silverhorn.LocalDataPath = "client/silverhorn.dat"

-- Set up default data
Silverhorn.LocalData["game"] = {
	-- Control if you've seen each of the endings
	has_ending_g = false,
	has_ending_b = false,
	has_ending_n = false,
	has_ending_t = false,
	
	emeralds_obtained = false, -- If you've obtained the emeralds legit at least once
	ms_accessed = false, -- If you've entered mindscape legit using the portal
	
	--If you cleared mindscape at least once
	--NOTE: This controls if you can use the silverhorn command! (INAZUMA.lua)
	ms_clear = false, 
}

Silverhorn.LocalData["setting"] = {
	--Lightning color data (LIGHTNINGCOLOR.lua)
	lightningcolor1 = "None",
	lightningcolor2 = "None",
	
	--Dance button data (SilverhornDanceButton.lua)
	--Make sure this matches DEFAULTVALUE in that Lua!
	dancebutton1 = BT_CUSTOM2, 
	dancebutton2 = BT_CUSTOM2,
	
	aetherhud_disabled = false, --Disabled Aether's hud (AetherHUD.lua)
	silverhorn_enabled = false, --Enabled Inazuma's silver horns (INAZUMA.lua)
}
-- TODO: deleteme
function Silverhorn.CompareGameValue(checkname, ischecknameval, cond, newvalue, altvalue)
	if ((Silverhorn.LocalData["game"][checkname] == ischecknameval) and cond) then
		return newvalue
	else
		return altvalue
	end
end
-- TODO: deleteme
function Silverhorn.CheckGameValue(checkname, isval, cond)
	return ((Silverhorn.LocalData["game"][checkname] == isval) and cond) and true or false
end

-- Load the save data here, since we only need to do it once for the local player
-- Essentially, we iterate through each of Silverhorn.LocalData's values and apply
-- the equivalent save value to it, with any extra values in save being ignored
local save = Silverhorn.LoadSave()
if save
	-- Iterate through each section
	for s, fields in pairs(Silverhorn.LocalData)
		-- No equivalent section? Something must've went wrong!
		if save[s] == nil
			print("\x83 WARNING:\x80 Silverhorn Save Data is either missing or corrupt.")
		else
			-- Iterate through each field
			for f, data in pairs(fields)
				Silverhorn.LocalData[s][f] = save[s][f]
			end
		end
	end
else
	print("\x83 WARNING:\x80 Silverhorn Save Data is either missing or corrupt.")
end

// Lach: write save data to our file upon quitting the game
addHook("GameQuit", function(quit)
	if quit then
		Silverhorn.WriteSave()
	end
end)
