local Silverhorn = Silverhorn
local Inazuma = Silverhorn.Inazuma
local Aether = Silverhorn.Aether

Silverhorn.DEBUGPREFIX = "\140Silverhorn:\128 "

Silverhorn.debug = {value = 0} /*CV_RegisterVar(
{
	name = "silverhorn_debug",
	PossibleValue = {Off = 0, Logic = 1, Performance = 2, All = 3},
	defaultvalue = "Off"
})*/

local LARGE = "left"
local SMALL = "small"
local COLOR_CODES = {
	WHITE =		'\128',
	MAGENTA =	'\129',
	YELLOW =	'\130',
	GREEN =		'\131',
	BLUE =		'\132',
	RED =		'\133',
	GRAY =		'\134',
	ORANGE =	'\135',
	SKY =		'\136',
	PURPLE =	'\137',
	AQUA =		'\138',
	PERIDOT =	'\139',
	AZURE =		'\140',
	BROWN =		'\141',
	ROSY =		'\142',
	INVERTED =	'\143'
}

local ENDING_COLORS = {
	g = COLOR_CODES.GREEN,
	b = COLOR_CODES.ORANGE,
	n = COLOR_CODES.BLUE,
	t = COLOR_CODES.PURPLE
}

local statMetatable = {
	__index = {
		active = false,
		startTime = 0,
		time = 0,
		calls = 0,
		start = function(this)
			if this.active
				return
			end
			this.active = true
			this.calls = $ + 1
			this.startTime = getTimeMicros()
		end,
		stop = function(this)
			if not this.active
				return
			end
			this.active = false
			this.time = $ + getTimeMicros() - this.startTime
		end,
		reset = function(this)
			if this.active
				return
			end
			this.time = 0
			this.calls = 0
		end
	}
}

Silverhorn.perfStats = setmetatable({},
	{
		__usedindex = function(this, key)
			Silverhorn.Print("Performance stats tag " .. key .. " is being used already.")
		end,
		__index = setmetatable(
			{
				start = function(this, key)
					return this[key]:start()
				end,
				stop = function(this, key)
					return this[key]:stop()
				end,
				reset = function(this, key)
					return this[key]:reset()
				end
			},
			{
				__index = function(this, key)
					local stats = setmetatable({}, statMetatable)
					Silverhorn.perfStats[key] = stats
					return stats
				end
			}
		)
	}
)

local function ColorBool(bool)
	return bool and COLOR_CODES.GREEN or COLOR_CODES.RED
end

local function ColorEndings(str)
	local i = 1
	while i <= str:len()
		local char = str:sub(i, i)
		local color = ENDING_COLORS[char]
		
		if color ~= nil
			str = str:sub(1, i - 1) .. color .. char .. COLOR_CODES.WHITE .. str:sub(i + 1)
			i = $ + 3
		else
			i = $ + 1
		end
	end
	return str
end

function Silverhorn.Print(...)
	local str = ...
	if str == nil
		return
	end
	
	print(Silverhorn.DEBUGPREFIX .. str)
	Silverhorn.Print(select(2, ...))
end

function Silverhorn.DebugPrint(...)
	if Silverhorn.debug.value
		Silverhorn.Print(...)
	end
end


local step
local v, x, y, flags

local function DrawString(str, extraFlags, alignment)
	local step = step
	extraFlags = $ or 0
	alignment = $ or SMALL
	
	if str ~= nil
		v.drawString(x, y, str, flags|extraFlags, alignment)
	end
	
	if not alignment:find(SMALL)
		step = $ * 2
	end
	y = $ + step
end

function Inazuma.DrawDebug(v, player)
	local inazuma = player.inazuma
	
	if not inazuma
		return
	end
	
	DrawString("Player "..(#player + 1), 0, LARGE)
	DrawString()
	
	DrawString("Flags:")
	for flag, value in pairs(Inazuma.FLAGS)
		if flag:find("MASK")
			continue
		end
		
		DrawString(flag, (inazuma.flags & value) and V_GREENMAP or V_REDMAP)
	end
	DrawString()
	DrawString(string.format("Current normalspeed: %f", player.normalspeed))
	DrawString(string.format("Stored normalspeed: %f", inazuma.normalspeed))
	DrawString(string.format("Blaze hack: %s%s", ColorBool(player.blazespindash), tostring(player.blazespindash)))
	DrawString(nil, 0, LARGE)
end

function Aether.DrawDebug(v, player)
	local aether = player.aether
	
	if not aether
		return
	end
	
	DrawString("Player "..(#player + 1), 0, LARGE)
	DrawString()
	
	DrawString("Flags:")
	for flag, value in pairs(Aether.FLAGS)
		if flag:find("MASK")
			continue
		end
		
		DrawString(flag, (aether.flags & value) and V_GREENMAP or V_REDMAP)
	end
	DrawString()
	DrawString(string.format("Wings table: %s%s", ColorBool(aether.wings), tostring(aether.wings ~= nil)))
	DrawString()
	DrawString(string.format("Current normalspeed: %f", player.normalspeed))
	DrawString(nil, 0, LARGE)
end

function Silverhorn.DebugHUD(drawer, player)
	local debug = Silverhorn.debug.value
	
	if not debug
	or player ~= displayplayer
		return
	end
	
	v = drawer
	
	if debug & 1
		x = 16
		y = 56
		step = 4
		flags = V_SNAPTOLEFT|V_SNAPTOTOP
		
		DrawString("Server-side endings unlocked: " .. ColorEndings(Silverhorn.GetPlayerEndingsString(player)))
		if consoleplayer == displayplayer
			DrawString("Client-side endings unlocked: " .. ColorEndings(Silverhorn.GetLocalEndingsString()))
		end
		DrawString()
		
		for player in players.iterate
			if player ~= displayplayer
			and player ~= secondarydisplayplayer
			and player.botleader ~= displayplayer
				continue
			end
			
			Inazuma.DrawDebug(v, player)
			Aether.DrawDebug(v, player)
		end
	end
	
	if debug & 2
		x = 316
		y = 184
		step = -4
		flags = V_SNAPTOLEFT|V_SNAPTORIGHT|V_MONOSPACE|V_ALLOWLOWERCASE
		
		for key, stat in pairs(Silverhorn.perfStats)
			DrawString(key .. ": " .. (" "):rep(5 - string.len(stat.time)) .. stat.time .. ", calls: " .. (" "):rep(3 - string.len(stat.calls)) .. stat.calls, 0, "small-right")
		end
	end
end

addHook("PreThinkFrame", function()
	local perfStats = Silverhorn.perfStats
	
	perfStats:reset("perfStats reset")
	perfStats:start("perfStats reset")
	
	for _, stat in pairs(perfStats)
		stat:reset()
	end
	
	perfStats:stop("perfStats reset")
end)

hud.add(Silverhorn.DebugHUD)