local ENDING_G = 1 << 0
local ENDING_T = 1 << 1
local ENDING_N = 1 << 2
local ENDING_B = 1 << 3

local endingFlags = {
	has_ending_g = ENDING_G,
	has_ending_t = ENDING_T,
	has_ending_n = ENDING_N,
	has_ending_b = ENDING_B
}

// returns whether @player has unlocked @ending
function Silverhorn.PlayerHasEnding(player, ending)
	local endings = Silverhorn.GetPlayerEndingFlags(player, flags)
	local flag = endingFlags[ending]
	
	assert(flag, "Unknown ending type \"" .. ending .. "\"")
	
	if endings == nil
		return false
	end
	
	return (endings & flag)
end

// returns the unlocked endings from the local savedata as a bitfield
function Silverhorn.GetLocalEndingFlags()
	local data = Silverhorn.LocalData["game"]
	local flags = 0
	
	for ending, flag in pairs(endingFlags)
		if data[ending]
			flags = $ | flag
		end
	end
	
	return flags
end

// sets @player's unlocked endings using a bitfield, like the one returned by Silverhorn.GetLocalEndingFlags()
function Silverhorn.SetPlayerEndingFlags(player, flags)
	if player == secondarydisplayplayer
		player = consoleplayer
	end
	
	player.silverhorn_endings = flags
end

// returns @player's unlocked endings as a bitfield, like the one returned by Silverhorn.GetLocalEndingFlags()
function Silverhorn.GetPlayerEndingFlags(player)
	if player == secondarydisplayplayer
		player = consoleplayer
	end
	
	return player.silverhorn_endings
end

// unlocks @ending for all players and saves it to the local savedata
function Silverhorn.AchieveEnding(ending)
	local flag = endingFlags[ending]
	
	assert(flag, "Unknown ending type \"" .. ending .. "\"")
	
	for player in players.iterate
		if player == secondarydisplayplayer
			continue
		end
		
		local endings = Silverhorn.GetPlayerEndingFlags(player)
		
		if endings == nil // uh oh, we haven't received this player's endings yet. we'll have to add them later, once they're received.
			player.silverhorn_pendings = ($ or 0) | flag
		else
			Silverhorn.SetPlayerEndingFlags(player, endings | flag)
		end
	end
	
	Silverhorn.SaveValue("game", ending, true)
end

Silverhorn.AddExecCallback(function(player, playerNum)
	if playerNum ~= 1
	or Silverhorn.GetPlayerEndingFlags(player) ~= nil
		return
	end
	
	CommandQueue_Add("_silverhorn_sendgamedata " .. Silverhorn.GetLocalEndingFlags())
end)

Silverhorn.AddPlayerConnectCallback(function(player)
	Silverhorn.SetPlayerEndingFlags(player, nil)
end)

COM_AddCommand("_silverhorn_sendgamedata", function(player, arg)
	if arg == nil
	or Silverhorn.GetPlayerEndingFlags(player) ~= nil
		return
	end
	
	arg = tonumber($)
	
	if arg == nil
		return
	end
	
	if player.silverhorn_pendings // add any endings achieved between joining the game and receiving the netcommand
		arg = $ | player.silverhorn_pendings
		player.silverhorn_pendings = nil
	end
	
	Silverhorn.SetPlayerEndingFlags(player, arg)
end)

// for debugging purposes

function Silverhorn.GetPlayerEndingsString(player)
	local str = ""
	local endings = Silverhorn.GetPlayerEndingFlags(player)
	
	if endings == nil
		return str
	end
	
	for ending, flag in pairs(endingFlags)
		if endings & flag
			str = $ .. ending:sub(ending:len())
		end
	end
	
	return str
end

function Silverhorn.GetLocalEndingsString()
	local str = ""
	local data = Silverhorn.LocalData["game"]
	
	for ending, _ in pairs(endingFlags)
		if data[ending]
			str = $ .. ending:sub(ending:len())
		end
	end
	
	return str
end
